%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 370-ANYREG.SL - patterns and functions for IBM370 PSL anyreg expansion   %
%                                                                          %
% Author:      Jim McDonald                                                %
%              IMSSS                                                       %
%              Stanford University                                         %
% Date:        11 December 1982                                            %
% Based on HP-CMAC.SL of Mar 82 by Maguire, Griss, Lowder                  %
% Based on VAX-CMAC.SL of 14 Jan 1982 by Eric Benson                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       ANYREG DEFINITIONS                                  %
%                                                                           %
% These are the patterns for ANYREG definitions.                            %
% They are of the form:                                                     %
%                                                                           %
% DefAnyreg AnyregName ResolutionFunction                                   %
%           ((RegisterPredicate1) (Returned value))                         %
%           ((RegisterPredicate2) (returned value))                         %
%                     :                    :                                %
%                    etc.                                                   %
%                                                                           %
% The last element in a defintion has no predicate--it is the fall-through  %
%  (default) case.                                                          %
%                                                                           %
% AnyregName is flagged (on the property list) with 'Anyreg, and has two    %
% properties:                                                               %
%                                                                           %
%   AnyregResolutionFunction -- function called to resolve the Anyreg       %
%   AnyregPatternTable       -- pairs of predicates and returned values     %
%                                                                           %
% The associated functions of most ANYREG's are called with two arguments:  %
%  a temporary REGISTER, used in the computation if needed,                 %
%  and the single SOURCE, acting as an extended addressing mode.            %
%                                                                           %
% The associated function usually uses                                      %
%    (OneOperandAnyreg Register Source 'Anyregname) or                      %
%    (ExpandOneArgumentAnyReg Register Source 'AnyregName),                 %
%    but it may be arbitary code.                                           %
%                                                                           %
%    (OneOperandAnyreg Register Source Anyregname) calls the lower level    %
%    ExpandOneArgumentAnyreg, after replacing Source by                     %
%    (ResolveOperand Register Source), which tries to reduce expressions    %
%    that can be simplified or evaluated at compile-time.                   %
%                                                                           %
% The MEMORY anyreg means that the location is the address of the source or %
%    destination.  This anyreg has an additional argument, called ArgTwo.   %
%                                                                           %
% Note:  It is important that the anyregs DO NOT assign any explicit        %
%        registers, because an instruction may have two similar anyregs     %
%        which would cause the explicit register to be used twice in the    %
%        same instruction.                                                  %
%                                                                           %
% The CMACRO !*Fetch can be used to load a register with a byte, halfword,  %
% or fullword which is directly addressible.  It chooses the appropriate    %
% load instructions, and uses only the given register.                      %
% Nov. 1, 1983 -- It now also works for big offsets.                        %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefAnyreg CAR AnyregCAR     %grab the source so caller can displace off it.
        ((RegP)                (indexed Source 0))
        % could optimize car of constant,
        % but car of a constant is wierd, probably an error
        (                      (!*Fetch Register Source)
                               (indexed Register 0)))

(DefAnyreg CDR AnyregCDR     %Same as CAR, except move to next word in pair.
        ((RegP)                (indexed Source 4))
        % could optimize cdr of constant,
        % but cdr of a constant is wierd, probably an error
        (                      (!*Fetch Register Source)
                               (indexed Register 4)))

(DefAnyreg Frame AnyregFrame
        ((DispInumP)           (indexed (Reg CtlStkR) Source))
        % We could move other things into the Register and doubleindex,
        %  but a frame not specified by a number between 0 and 4096 is
        %  presumed to be an error.  I.e., more than 1024 local vars is
        %  presumed to be an error.
        (                      (ErrorFrame Source)))

(De AnyregImmediate (Register Source)
   (cond ((InumP Source)              Source)
         ((Eqcar Source 'Unimmediate) Source)
         (T                           (list 'Immediate Source))))

(DefAnyreg Immediate AnyregImmediate)

(DefAnyreg Memory AnyregMemory          % c(source) + argtwo
        % c(register) + immediate displacement
        ((RegP      DispInumP)  (Indexed Source ArgTwo))  % Disp includes 0

        % Warning -- LA zeros out top byte of result, so negative offsets
        %            cannot be created via LA Ry,0(Rx,Ry)
        % The WPlus2, while technically illegal here, will work since
        %  adding and subtracting Inums never involves extra registers.
        % In some cases, this may yield inefficient code,
        %  since LA Ry,0(Ry) may result (see anyreg Address below)
        ((RegP      NegFullInumP)  (LR Register Source)
                                   (SameLineComment "Inefficient sequence")
                                   (!*WPlus2 Register ArgTwo)
                                   (Indexed Register 0))
        % WPlus2 may generate this
        ((RegP      FullInumP)  (!*Fetch Register ArgTwo)
                                (DoubleIndexed Source Register 0))

        % c(reg) + c(reg)       e.g. (Memory (Reg 1) (WArray SymFnc))
        ((RegP      RegP)       (DoubleIndexed Source ArgTwo 0))

        % c(register) + memory location
        ((RegP      WArrayP)    (L Register ArgTwo)     % indirection
                                (SameLineComment Register ":= base of warray"
                                                 ArgTwo)
                                (DoubleIndexed Source Register 0))
        ((RegP      AnyP)       (ErrorMemory Source ArgTwo Register))

        % c(memory location) + immediate displacement
        ((AnyP      DispInumP)  (!*Fetch Register Source)
                                (Indexed Register ArgTwo))
        ((AnyP      HalfInumP)  (!*Fetch Register Source)
                                (AH Register (Lit (Halfword ArgTwo)))
                                (Indexed Register 0))
        ((AnyP      FullInumP)  (!*Fetch Register Source)
                                (A Register (Lit (Fullword ArgTwo)))
                                (Indexed Register 0))

        (                       (ErrorMemory Source ArgTwo Register)))

%
%  ADDRESS should be just like MEMORY, except that final object is
%  prefixed by IMMEDIATE.
%
(DefAnyreg Address AnyregAddress)
(de AnyregAddress (Register Source ArgTwo)
 (list 'Immediate (TwoOperandAnyreg Register Source ArgTwo 'MEMORY)))

(DefCMacro !*Fetch)
% NOTE:  !*Fetch is used by anyreg's.
%        It must NOT use ANY registers except those in Reg and Source.

(de !*Fetch (Reg Source)
 (nconc
  (cond ((ByteInumP Source)
         (list (list 'SR Reg Reg)
               (list 'SameLineComment "Clear" Reg)
               (list 'IC Reg (list 'Lit (list 'Byte Source)))))
        ((HalfInumP Source)
         (list (list 'LH Reg (list 'Lit (list 'Halfword Source)))))
        ((FullInumP Source)
         (list (list 'L  Reg (list 'Lit (list 'Fullword Source)))))
        ((ByteP Source)
         (list (list 'SR Reg Reg)
               (list 'SameLineComment "Clear" Reg)
               (list 'IC Reg Source)))
        ((HalfwordP Source)
         (list (list 'LH Reg Source)))
        ((BigOffsetP Source)
         (list (list '!*Move (list 'OffsetOf Source) Reg)
               %  this move will be a L or LH of a literal
               %  no extraneous regs are involved (see warning above)
               (list 'SameLineComment Reg ":= offset of" Source)
               (list 'L Reg
                     (list 'DoubleIndexed Reg (list 'BaseRegOf Source) 0))))
        ((FullwordP Source)
         (list (list 'L  Reg Source)))
        (t
         (list (list 'ErrorFetch Reg Source))))
  (list (list 'SameLineComment Reg ":=" Source))))

(DefAnyreg Quote AnyregQuote
        ((Null)          (Reg NilR))       % If NIL, return the NIL register
        ((EqTP)          (fluid T))        % ??? Is this defined?
        ((InumP)         Source)           % ?
        (                (Quote Source)))  % ? Recursivly expand??

(DefAnyreg Reg AnyregReg
        ((FakeRegP)      (ExtraReg Source))
        (                (Reg Source)))

(DefAnyreg WVar AnyregWVar)
(DefAnyreg WArray AnyregWVar)

(de AnyRegWVar (Register Source)
 (prog (Alias)
       (return
        (cond ((RegisterNameP Source) (list 'Reg Source))
              ((setq Alias (Get Source 'RegisterAlias)) (list 'Reg Alias))
              (t (list 'WVar Source))))))

(DefAnyreg OffsetOf AnyregOffsetOf)

(de AnyRegOffsetOf (Register Source)
 (cond ((atom source) (list 'OffsetOf Source))
       ((memq (car source) '(FnCell Fluid Global !$Fluid !$Global))
        % (times AddressingUnitsPerItem (FindIdNumber (cadr Source))))
        (list 'IdOffset (cadr Source)))
       (t (list 'OffsetOf Source))))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       end of file                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
