%***********************************************************************
%
% (C) Copyright 1989
% File:             h-help.red
% Object File:      help
%
% Basic help functions. The following non-Standard Lisp functions are
% required:
%
%   readchar()   - Reads ASCII character from random access file. All
%                  help and formatting input go through this function so
%                  it is not necessary to be part of the Lisp I/O system.
%   ropen(file)  - Open "file" for random access input.
%   rseek(n)     - Move the input pointer to character n in the currently
%                  active random access input file.
%   rrds(fileh)  - Select the random access file for input or return to
%                  standard input. This is a hack to make UO-LISP work
%                  and most systems should just make this RDS.
%   #RFILEH      - A UO-LISPism used by the random access I/O package.
%
% We assume that each line ends with <cr> <lf>.
%
% Revision History: (Created 10/10/89)
%   12/18/89 Switch to list, multiple-topic form, add parsing of help.
%   12/24/89 Add h!/restart for session freeze.
%   1/21/90 Flag HELP as a GO to allow HELP.
%   12/1/91 Parse HELP; as h/error.
%   12/20/91 Do the RRDS function for other systems, add error hacks.
%
%***********************************************************************


% Channel for random access help file.
global '(!#RFILEH);

% Here are the fields kept around during processing.
fluid '(NEXT!* PREVIOUS!* TOP!* RELATED!* RELATEDPTR!* HTABLE!* HFNAME!*);

% Globals imported from the formatter.
global '(h!:tok h!:eof h!:curc);

expr procedure h!/init(fhlp, fidx);
% fhlp is the name of the help file, fidx is its index. Read the index file
% and make all the topics into pseudo help functions. 
begin scalar x, finh, olds, topic;
  finh := rds open(fidx, 'input);
  HTABLE!* := read();
  close rds finh;
  for each tp in HTABLE!*
      do << put(car tp, 'STAT, 'hparse);     % Make into a stat.
            flag({car tp}, 'NOFORM) >>;   % Don't fiddle it.
  !#RFILEH := ropen (HFNAME!* := fhlp)       % Returns a DOS handle.
end;

expr procedure h!/restart; !#RFILEH := ropen HFNAME!*;

expr procedure hparse;
% Parse a help statement which is a bunch of symbols up to a semicolon or
% dollar sign.
begin scalar topics, fn;
  fn := cursym!*;
loop: scan();
  if cursym!* eq '!*semicol!* then
    return if null topics then {'h!/error}
           else if fn eq 'HELP then
                {'h!/topics, 'htable!*, mkquote reversip topics}
           else {'h!/topics, mkquote {atsoc(fn, htable!*)},
                 mkquote reversip topics};
  topics := cursym!* . topics;
  scan();
  if cursym!* eq '!*comma!* then go to loop;
  if cursym!* eq '!*semicol!* then
   return if fn eq 'HELP then
            {'h!/topics, 'htable!*, mkquote reversip topics}
       else {'h!/topics, mkquote {atsoc(fn, htable!*)},
                 mkquote reversip topics}
    else error(0, "missing comma in HELP list");
end;
put('help, 'stat, 'hparse);
flag('(help), 'noform);
flag('(help), 'go);

expr procedure h!/topics(tpc, x);
% tpc is a list of topic alists. x is a list of things to look for in there.
% Put all the things we can find on the RELATED!* list and scan the first one.
begin scalar loc;
  RELATED!* := RELATEDPTR!* := nil;
  for each tp in tpc
      do for each tpx in x 
         do if loc := atsoc(tpx, tp) then RELATED!* := loc . RELATED!*;
  if null RELATED!* then
     return << prin2 "No help available on "; print x >>;
  RELATEDPTR!* := cdr RELATED!*;
  h!/topic1 cdar RELATED!*
end;

fluid '(h!:enders);                      % Imported from h-format.red

expr procedure h!/topic1 loc;
% locs is a single character pointer or a list of character pointers.
begin scalar h!:enders, x, oldfh;
  oldfh := rrds !#RFILEH;
loopa: if not izerop(x := rseek loc) then
     return << rrds oldfh;
               prin2 "Internal help error "; prin1 x; prin2t "on seek" >>;
  h!:curc := nil;
  h!:tok := h!/ntok();
loop: if h!/comm t then if not(h!:tok eq h!:eof) then go to loop;
  rrds oldfh;
  return t
end; 

expr procedure h!/next;
% Move to the next topic in the help.
 if numberp NEXT!* then h!/topic1 NEXT!*
  else << prin2t "*** There is no next help topic."; nil >>;

expr procedure h!/previous;
% Move to the previous topic in the help system.
if numberp PREVIOUS!* then h!/topic1 PREVIOUS!*
  else << prin2t "*** There is no previous help topic."; nil >>;

expr procedure h!/top;
% Move to the top level help topic in the help system.
if numberp TOP!* then h!/topic1 TOP!*
  else << prin2t "*** There is no top level help topic."; nil >>;

expr procedure h!/related;
% Move down the related list. 
  if RELATEDPTR!* then
     << if pairp car RELATEDPTR!* and numberp cdar RELATEDPTR!* then 
            h!/topic1 cdar RELATEDPTR!*
          else << prin2 "*** No help on ";
                  prin1 car RELATEDPTR!*;
                  prin2t ": try RELATED; again!"; nil >>;
        RELATEDPTR!* := cdr RELATEDPTR!*; t >>
    else << prin2t "*** No more related help!"; nil >>;


%-----------------------------------------------------------------------
% Error processing
%-----------------------------------------------------------------------

global '(LASTPNAME!* LASTNUMBER!* LASTMESSAGE!*);

% Temporary hack since we know what we're doing.
remprop('rerror, 'formfn);

symbolic procedure rerror(packagename, number, message);
% This replaces the rerror in RLISP. We save the package name and number
% for this error so that we can get at it when the user asks for help.
<< LASTPNAME!* := packagename;
   LASTNUMBER!* := number;
   LASTMESSAGE!* := message;
   rederr message >>;


symbolic procedure h!/error;
% We have the user entering HELP, no arguments probably after an error.
% If no error, then default to help help; otherwise help on the error
% number from RERROR. Before the error message try and print as much as is
% known about the error.
if null LASTPNAME!* then {'h!/topics, 'HTABLE!*, '(HELP)}
else begin scalar tpn;
   terpri();
   prin2t "The last known REDUCE error:";
   lprie LASTMESSAGE!*;
   terpri();
   if null(tpn := atsoc(LASTPNAME!*, HTABLE!*)) then
      << prin2 "Unfortunately, there is no help available for the package ";
         prin1 LASTPNAME!*;
         prin2 ", with error number "; print LASTNUMBER!* >>
     else h!/topics({cdr tpn}, {LASTNUMBER!*})
  end;


%-----------------------------------------------------------------------
% Hack to REDUCE 3.4 to make symerr do something reasonable. Try and
% convert the error messages into a number for the HELP system.
%-----------------------------------------------------------------------

global '(!3!4NIL !3!4T);

!3!4NIL := '(
             ("Invalid parameter" . (RLISP . 30))
             ("Syntax error: improper number" . (RLISP . 31))
             ("Invalid S-expression" . (RLISP . 32))
             ("Syntax error in procedure header" . (RLISP . 33))
             ("Improper delimiter" . (RLISP . 34))
             ("Too many right parentheses" . (RLISP . 35))
             ("Redundant operator" . (RLISP . 36))
             ("Improper delimiter" . (RLISP . 37))
             ("Missing operator" . (RLISP . 38))
             ("Improper delimiter" . (RLISP . 39))
             ("Too few right parentheses" . (RLISP . 40))
             ("Improper delimiter" . (RLISP . 41))
             ("Improper delimiter" . (RLISP . 42))
             ("Syntax error: semicolon in list" . (RLISP . 43))
             ("Syntax error: invalid comma in list" . (RLISP . 44))
             ("Syntax error" . (MATRIX . 25))
              );

!34T := '(
          (if . (RLISP . 50))
          ("Function" . (RLISP . 51))
          (NIL . (RLISP . 52))
          (block . (RLISP . 53))
          (procedure . (RLISP . 54))
          (for . (RLISP . 55))
          ("FOR EACH" . (RLISP . 56))
          (repeat . (RLISP . 57))
          (while . (RLISP . 58))
          (forall . (ALG . 35))
          (let . (ALG . 36))
          (coframe . (EXCALC . 11))
          (ghorner . (SCOPE . 9))
         );


symbolic procedure symerr(u,v);
% This version hacked to change funny error messages into real error
% numbers. Later symerr should have 2 extra arguments with the package
% and number added.
begin scalar x, pnn;
  erfg!* := t;
  if numberp cursym!* or not(x := get(cursym!*,'prtch))
    then x := cursym!*;
  terpri();
  if !*echo then terpri();
  outl!*:=car outl!* . '!$!$!$ . cdr outl!*;
  comm1 t;
  mapcar(reversip!* outl!*,function prin2);
  terpri();
  outl!* := nil;
  if null v then
     << pnn := assoc(u, !3!4NIL);
        if null pnn then rerror('rlisp, 5, u)
          else rerror(cadr pnn, cddr pnn, u) >>
   else
     << pnn := assoc(u, !3!4T);
        if null pnn then
          rerror('rlisp,5, x . ("invalid" .
                                if u then list("in",u,"statement")))
         else rerror(cadr pnn, cddr pnn,
                     x . ("invalid" . if u then list("in",u,"statement"))) >>
end;


% Make rerror special again.
put('rerror, 'formfn, 'formrerror);


end;


