%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:TOP-LOOP.SL
% Title:        Generalized top loop construct
% Author:       Eric Benson and M. L. Griss
% Created:      19 October 1981
% Modified:     23-Oct-84 14:51:52 (Alan Snyder)
% Package:
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(toploopread*    % reading function
     toploopprint*   % printing function
     toploopeval*    % evaluation function
     toploopname*    % short name to put in prompt
     toplooplevel*   % depth of top loop invocations
     promptstring*   % input prompt
     lispbanner*     % Welcome banner printed in StandardLisp
     *emsgp          % whether to print error messages
     *backtrace      % whether to print backtrace
     *time           % whether to print timing of evaluation
     gctime*         % Time spent in garbage collection
     *defn           % To "output" rather than process
     dfprint*        % Alternate DEFN print function
     *output         % Whether to print output.
     semic*          % Input terminator when in Rlisps.
     *nonil          % Whether to supress NIL value print.
     initforms*      % Forms to be evaluated at startup
     lispscantable*  % CurrentScanTable!* when READing
     *package*       % Current package
     *break          % Controls entry into debugger on error
     startuproutine*
     using-other-bps-spaces*
     ))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MAIN function and related stuff
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(loadtime (setf *redefmsg t))
 
(de pre-main ()
  (cleario)
  (clearbindings)
  (initializeinterrupts)
  (evalinitforms)
  (setq toplooplevel* -1)
  (setf toploopread*  'read)
  (setf toploopeval*  'eval)
  (setf toploopprint* 'print)
  (history-reset)
  (when (null lispbanner*)
    (setq lispbanner* "Portable Standard LISP"))
  (setq *output t)    % Output ON by default.
  (when (equal (catch 'reset (main)) 'reset)
    (pre-main)
    ))
 
(de main ()
  %
  % A new system can be created by redefining this function to call whatever
  % top loop is desired.
  %
  (setq startuproutine* 'standardlisp)
  (apply startuproutine* 'nil)
  )
 
(de reset ()
  (throw 'reset 'reset)
  )
 
(de standardlisp ()
  % Lisp top loop
  ((lambda (currentreadmacroindicator* currentscantable*)
           (toploop 'read 'printwithfreshline 'eval "lisp" lispbanner*))
   'lispreadmacro
   lispscantable*))
 
(de savesystem (banner file initforms)
  (when using-other-bps-spaces*
    (stderror "Cannot unexec after allocating additional bps space"))
  (let ((history (history-state)))
    (history-reset)
    (setq lispbanner* (bldmsg "%w, %w" banner (date)))
    (setq *usermode t)
    (setq initforms* initforms)
    (dumplisp file)
    (setq initforms* nil)
    (set-history-state history)
    NIL
    ))
 
(de evalinitforms ()
  % Evaluate and clear InitForms*
  (for (in x initforms*)
       (do (eval x)))
  (setq initforms* nil)
  )
 
(de time () % why is this here???
  % Get run-time in milliseconds
  (sys2int (timc))) % TimC is primitive runtime function
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TOPLOOP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de toploop (toploopread* toploopprint* toploopeval* toploopname* welcomebanner)
  % Generalized top-loop mechanism
  (let ((toplooplevel* (add1 toplooplevel*)) % FLUID
    (semic* ';) % FLUID -- Take out and put in Rlisp????
    (promptstring* promptstring*) % preserve this FLUID variable
    )
    (prin2 welcomebanner)
    (terpri)
    (while t
      (let ((count (history-append)))
(adb 10)
    (setq promptstring* (compute-prompt-string count toplooplevel*)))
(adb 11)
      (let ((input-value (top-loop-read)))
    (when (or (eq input-value '$exittoploop$) (eq input-value $eof$))
      (exit))
    (history-set-input input-value)
    (cond ((not *defn) (top-loop-eval-and-print input-value)) % normal case
          (dfprint* (funcall dfprint* input-value)) % "Compile"
          (t (prettyprint input-value))
          ))))
  (printf "Exiting %w%n" toploopname*)
  )
 
(de top-loop-read ()
  (while t
    (let ((input-value (errset (funcall toploopread*) t)))
      (when (pairp input-value) (exit (car input-value))))))
 
(de top-loop-eval-and-print (input-value)
  (let* ((this-gc-time gctime*)
    (time-check (time))
    (output-value (errset (funcall toploopeval* input-value) t))
    )
    (setq time-check (difference (time) time-check))
    (setq this-gc-time (difference gctime* this-gc-time))
    (when (pairp output-value)
      (setq output-value (car output-value))
      (history-set-output output-value)
      (when (and *output (eq semic* ';)
         (not (and *nonil (eq output-value nil))))
    (errset (funcall toploopprint* output-value) t))
      (when *time
    (if (equal this-gc-time 0)
      (printf "Cpu time: %w ms%n" time-check)
      (printf "Cpu time: %w ms, GC time: %w ms%n"
          (difference time-check this-gc-time) this-gc-time)))
      )))
 
(de printwithfreshline (x)
  (printf "%f%p%n" x))
 
(de compute-prompt-string (count level)
  (mybldmsg "%w %w%w%w%w%w "
      count
      (if *defn "*DEFN " "")
      toploopname*
      (if (igreaterp level 0) (mybldmsg " (%w)" level) "")
      (if *package* (mybldmsg " [%w:]" (eval '(package-name *package*))) "")
      (if (or (igreaterp level 0) *package*) " >" ">")
      ))
 
(de mybldmsg(f a1 a2 a3 a4)
   (prog (x)
      (adb 65)
      (setq x (bldmsg f a1 a2 a3 a4))
      (adb 66)
      (return x)))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The HISTORY Mechanism
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(historycount*   % number of entries read so far
     historylist*    % list of entries read and evaluated
     ))
 
(de history-state ()
  % Return a description of the current History.
  (list historycount* historylist*))
 
(de history-reset ()
  (setq historylist* nil)
  (setq historycount* 0)
  )
 
(de set-history-state (x)
  (setq historycount* (first x))
  (setq historylist* (second x))
  )
 
(de history-append ()
  % Add a new entry and return its number.
  (setq historycount* (add1 historycount*))
  (setq historylist* (cons (cons nil nil) historylist*))
  historycount*
  )
 
(de history-set-input (value)
  (rplaca (car historylist*) value))
 
(de history-set-output (value)
  (rplacd (car historylist*) value))
 
(de history-fetch (n)
  % Return the Nth entry in the History list.
  (let ((*break NIL))
    (if (or (ilessp n 0) (igeq n historycount*))
      (stderror (bldmsg "No history entry %r" n))
      (car (pnth (cdr historylist*) (idifference historycount* n))))))
 
(de inp (n)
  % Return Nth input
  (car (history-fetch n)))
 
(de redo (n)
  % Re-evaluate Nth input
  (apply toploopeval* (list (car (history-fetch n)))))
 
(de ans (n)
  % Return Nth output
  (cdr (history-fetch n)))
 
(dn hist (al)
  % Print history entries:
 
  % (HIST 'CLEAR) => resets the history list
  % (HIST -n) => print the last N entries
  % (HIST n) => print entries starting with #N
  % (HIST m n) => print entries #M through #N
 
  (unless (ilessp historycount* 2)
    (cond ((and (pairp al) (equal (car al) 'clear))
       (history-reset)
       (history-append)
       (history-set-input '(hist clear))
       nil)
      ((and (pairp al) (iminusp (car al)))
       (histprint (cdr historylist*) (isub1 historycount*)
              (iminus (car al))))
      (t
       (let ((i1 1) (i2 (isub1 historycount*)))
         (when (pairp al)
           (setq i1 (max i1 (car al)))
           (setq al (cdr al)))
         (when (pairp al)
           (setq i2 (min i2 (car al))))
         (histprint
          (pnth (cdr historylist*) (idifference historycount* i2))
          i2
          (iadd1 (idifference i2 i1))))))))
 
(de histprint (l n m)
  (unless (izerop m)
    (histprint (cdr l) (isub1 n) (isub1 m))
    (printf "%w%tInp: %p%n%tAns: %p%n" n 5 (car (first l)) 5 (cdr (first l)))
    ))
 
% End of file.
