%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxu/symget.sl (DEC Alpha version)
% Description:    fast access to 'first class' properties. 
%                 used for reducing the number of propertylist requests
% Author:         Winfried Neun , ZIB Berlin
% Created:        13-February-1993
%
% Copyright (c) 1993, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Usage:
% Get requests asking for a 'first class' property are handled via a two
% step vector access: first a vector (length maxsymbols) is inspected for
% an entry of 'first class' properties. If this exists, the second vector
% containing the special properties is used. Otherwise NIL.
% The vector for all ids is crested in the kernel, the id specific second
% vector is created by put operation. Put also alters the property list to
% ensure correctness (e.g. in interpreter)
%
% For flags all property list activities are done via fast vector access.
%
% The property to be a 'first class' property must be set at the very
% early beginning of compilation
%

(load fast-vector) 
(on fast-vectors)

(compiletime (progn

 (setq *syslisp t)                         
 (load fast-vector) (on fast-vectors)
 (load if-system)

 (ds string-concat(a b)
       (compress (append (explode2 a) (explode2 b))))

 (defmacro def-pass-1-macro (name args . body)
   (let ((fcn-name (intern (string-concat "PA1M-" (id2string name)))))
     `(progn
       (put (quote ,name) (quote pass-1-macro) (quote ,fcn-name))
       (defmacro ,fcn-name ,args ,@body)
    )))

 (remprop 'get 'pass-1-macro)   % avoid endless recursion
))

(ds symgt (id prop)
       ((lambda (v propval)
	  (cond ((null v) NIL)
		(t (igetv v propval))))
       (symget (inf id)) prop))

(de install-symflag (x val)
       (put x 'symflag? t)
       (put x 'symflagval val))

(de install-symget (x val)       % install a first class property
       (when (memq x '(symget? symgetval)) (return nil))
       (put x 'symget? t)        % Caution cant symget id symget itself!
       (put x 'symgetval val))

(put 'symflag 'openfn  '(nonassocpat *symflag))

(de *symflag (ArgOne ArgTwo)
    (expand2operandcmacro ArgOne ArgTwo '*symflag))

(defcmacro *symflag  ((*move ($fluid &symflag&) (reg t4))
		      (*move argone (reg t5))
		      (zap (reg t5) 16#80 (reg t5))  % inf
		      (s8addq (reg t5) (reg t4) (reg t5)) % * 8
		      (zap (reg t5) 16#80 (reg t5)) % inf
		      (*move (displacement (reg t5) 8) (reg t5))
		      (*wshift (reg t5) (minus argtwo))
		      (*wand (reg t5) 1)
		      (*move (reg t5) argone)))

(compiletime
(progn 
  (put 'putsymflag 'opencode '((*move ($fluid &symflag&) (reg 5))
			       (sll (reg 1) 8 (reg 3)) % inf
			       (srl (reg 3) 5 (reg 3)) % *8
			       (*wplus2 (reg 5) (reg 3))
			       (zap (reg 5) 16#80 (reg 5)) % inf
			       (*move (displacement (reg 5) 8) (reg 4))
			       (*move 1 (reg 3))
			       (*wshift (reg 3) (reg 2))
			       (*wor  (reg 4) (reg 3))
			       (*move (reg 4) (displacement (reg 5) 8))))

  (put 'delsymflag 'opencode '((*move ($fluid &symflag&) (reg 5))
			       (sll (reg 1) 8 (reg 3)) % inf
			       (srl (reg 3) 5 (reg 3)) % * 8
			       (*wplus2 (reg 5) (reg 3))
			       (zap (reg 5) 16#80 (reg 5)) % inf
			       (*move (displacement (reg 5) 8) (reg 4))
			       (*move 1 (reg 3))
			       (*wshift (reg 3) (reg 2))
			       (*wnot (reg 3) (reg 3))
			       (*wand (reg 4) (reg 3))
			       (*move (reg 4) (displacement (reg 5) 8)))))
))%end compiletime

(dm symflagp (u)
  (cond ((not (get (cadr (caddr u)) 'symflag?)) nil)
      (t `(weq (symflag ,(cadr u) ,(get (cadr(caddr u)) 'symflagval))1)
      )))

(fluid '(&symflag& symget &symgetsize&))

(de !%put (a b c)
  (prog (v pu)
    (setq pu (old-put a b c))  % does checks
    (when (get b 'symget?)                      % if first class and
	  (setq v (symget (id2int a)))          % vector not present
	  (when (null v) (setq v (mkvect &symgetsize&))  % make it
			 (wputv  (inf symget) (id2int a) v))
	  (iputv v (get b 'symgetval) c))       % store property
    (return pu)))

(when (funboundp 'old-put) (copyd 'old-put 'put)
			   (copyd 'put '!%put))

(when (funboundp '!%get) (copyd '!%get 'get))
 
(de !%remprop (a b)
  (prog (v pu)
    (setq pu (old-remprop a b))  % does checks
    (when (get b 'symget?)
	  (setq v (symget (id2int a)))
	  (when v (iputv v (get b 'symgetval) NIL)))
    (return pu)))

(when (funboundp 'old-remprop) (copyd 'old-remprop 'remprop)
			       (copyd 'remprop '!%remprop))

(de !%flag1 (a b)
	 (when (get b 'symflag?) 
	       (putsymflag a (get b 'symflagval)))
	 (old-flag1 a b))

(when (funboundp 'old-flag1) (copyd 'old-flag1 'flag1)
			     (copyd 'flag1 '!%flag1))

(de !%remflag1 (a b)
	 (when (get b 'symflag?)
		 (delsymflag a (get b 'symflagval)))
	 (old-remflag1 a b))

(when (funboundp 'old-remflag1) (copyd 'old-remflag1 'remflag1)
				(copyd 'remflag1 '!%remflag1))

(null (progn
	     (setq &symgetsize& 32)                  % default
	     (setq &symflag& (make-words maxsymbols 0))))

(remob '!%put)
(remob '!%remprop)
(remob '!%flag1)
(remob '!%remflag1)

(def-pass-1-macro get (a b)
  (when (and (pairp b) (!%get (cadr b) 'symget?))
	  `((lambda (x)
	      (cond ((idp x) (symgt x ,(!%get (cadr b) 'symgetval)))
		    (t nil))) ,a )))

(def-pass-1-macro flagp (a b)
  (when (and (pairp b) (get (cadr b) 'symflag?))
	  `((lambda (x)
	       (cond ((and (idp x) (symflagp x ,b)) t)
		     (t nil))) ,a )))
