/*-------------------------------------------------------------------------*/
/*
 * File:         PXU:c2l.c
 * Description:  Routines to support C functions calling Lisp functions.
 * Author:       Harold Carr
 * Created:      August 1985
 * Package:
 * RCS $Header: $
 *
 * (c) Copyright 1987, University of Utah, all rights reserved
 */
/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/*
 * Usage: These depend on C having Lisp IDs (the bit pattern representing
 * a particular ID.  This can be accomplished in an initialization routine
 * called from Lisp giving it the ID of a LISP routine to call repeatly
 * for the values of IDs used by the C system, e.g.,:
 *
 * The Lisp part:
 *
 * (defvar s!%get-ids-counter!% 0
 *   "Successive calls on s!%get-ids!% will use this to
 *    decide which id to get.")
 *
 * % This list ***MUST*** be kept in the sync with list in C.
 *
 * (defvar s!%psl-ids-used-in-c!%
 *         '(s!%advance-time!%
 *           s!%bp-list!%
 *           s!%display-daemon-hook!%)
 *      "Ids used on C side to access PSL value cells and functions.")
 *
 *  (de s!%get-ids!% ()
 *     % This nth is PSL's, not CL's.
 *     (nth s!%psl-ids-used-in-c!% (incr s!%get-ids-counter!%)))
 *
 * The C part:
 *
 * int advanceTime;
 * int BPLIST;
 * int DISPLAYDAEMONHOOK;
 *
 * void
 * storeLispIdsInCvariables(getIds)
 * int getIds;
 * {
 *      advanceTime	  = call_lisp0(getIds);
 *      BPLIST 		  = call_lisp0(getIds);
 *      DISPLAYDAEMONHOOK = call_lisp0(getIds);
 *  }
 */
/*-------------------------------------------------------------------------*/

/* These two PSL symbols are the base of the value cell and function
   cell arrays.  You have to remove C's underscoring (_) of these two
   symbols before assembling this file. */

extern int SYMVAL;  /* These exist is the PSL system. */
extern int SYMFNC;

/* Put the base addresses in C globals. */

int simSymVal = (int) &SYMVAL;  /* Would rather just use the PSL globals */
int simSymFnc = (int) &SYMFNC;  /* directly in vca and fca macros below, */
                                /* but c compiler won't let me. */

#define ADDRESSINGUNITSPERVALUECELL 4
#define ADDRESSINGUNITSPERFUNCTIONCELL 6

#ifdef vax
#define TAGMASK 0xf8000000
#endif

#ifdef sun
#define TAGMASK 0xff000000
#endif

/* This calculates the value cell byte address. */

#define vca(id) (simSymVal + ((id & ~TAGMASK) * ADDRESSINGUNITSPERVALUECELL))

/* This calculates the function cell byte address. */

#define fca(id) (simSymFnc + ((id & ~TAGMASK) * ADDRESSINGUNITSPERFUNCTIONCELL))

/*-------------------------------------------------------------------------*/
/*
 * lisp_value_cell
 *
 * Given an id. Returns the contents of that id's value cell.
 */
/*-------------------------------------------------------------------------*/

int lisp_value_cell(id)
int id;
{
    return(*(int *)vca(id));
}

/*-------------------------------------------------------------------------*/
/*
 * set_lisp_value_cell
 *
 * Given an id and a value.
 * Sets the contents of that id's value cell to the given value.
 * Returns the given value.
 */
/*-------------------------------------------------------------------------*/

int set_lisp_value_cell(id, value)
int id, value;
{
    *(int *)vca(id) = value;
    return(value);
}

/*-------------------------------------------------------------------------*/
/*
 * run_lisp_func
 *
 * Given an address. Runs the code at that entry point.
 */
/*-------------------------------------------------------------------------*/

int run_lisp_func(address)
int (*address)();
{
#ifdef vax
    asm("jsb *4(ap)");   /* Use PSL calling convention. */
    asm("movl r1,r0");   /* Move PSL's return value to where C wants it. */
#endif

#ifdef sun
#ifdef DREGMODEL
    asm("movl a6,sp@-");     /* Save link reg because PSL clobbers it. */
    asm("movl a6@(0x8),a0"); /* Get address from argument stack. */
    asm("movl #0xfe000080,a6"); /* Set up NIL register. */
    asm("jsr  a0@");         /* Run func at that address. */
                             /* PSL result in d0, same as C. */
    asm("movl sp@,a6");      /* Restore link reg. */
    asm("addqw #4,sp");      /* Balance the stack. */
#endif
#ifdef AREGMODEL
    asm("movl a6,sp@-");     /* Save link reg because PSL clobbers it. */
    asm("movl a6@(0x8),a0"); /* Get address from argument stack. */
    asm("movl #0xfe000080,d0"); /* Set up NIL register. */
    asm("movl #0,a6");       /* Set up zeroareg. */
    asm("jsr  a0@");         /* Run func at that address. */
    asm("movl a1,d0");       /* Move PSL return value to where C expects it. */
    asm("movl sp@,a6");      /* Restore link reg. */
    asm("addqw #4,sp");      /* Balance the stack. */
#endif    
#endif
}

/*-------------------------------------------------------------------------*/
/*
 * call_lisp0, call_lisp1, ...
 *
 * These functions are given an id (and arguments).
 * They move the arguments into the registers as PSL expects
 * then runs the code in the ID's function cell.
 */
/*-------------------------------------------------------------------------*/

int call_lisp0(id)
int id;
{
    return(run_lisp_func(fca(id)));
}

int call_lisp1(id, arg)
int id, arg;
{
    int address = fca(id);/* Calculate address before moving args into regs. */

#ifdef vax
    asm("movl 8(ap),r1");
#endif

#ifdef sun
#ifdef DREGMODEL
    asm("movl a6@(12),d0");
#endif
#ifdef AREGMODEL
    asm("movl a6@(12),a1");
#endif    
#endif

    return(run_lisp_func(address));
}

int call_lisp2(id, arg1, arg2)
int id, arg1, arg2;
{
    int address = fca(id);

#ifdef vax
    asm("movl 12(ap),r2");
    asm("movl  8(ap),r1");
#endif

#ifdef sun
#ifdef DREGMODEL
    asm("movl a6@(16),d1"); 
    asm("movl a6@(12),d0");
#endif
#ifdef AREGMODEL
    asm("movl a6@(16),a2"); 
    asm("movl a6@(12),a1");
#endif
#endif

    return(run_lisp_func(address));
}

int call_lisp3(id, arg1, arg2, arg3)
int id, arg1, arg2, arg3;
{
    int address = fca(id);

#ifdef vax
    asm("movl 16(ap),r3");
    asm("movl 12(ap),r2");
    asm("movl  8(ap),r1");
#endif

#ifdef sun
#ifdef DREGMODEL
    asm("movl a6@(20),d2"); 
    asm("movl a6@(16),d1"); 
    asm("movl a6@(12),d0");
#endif
#ifdef AREGMODEL
    asm("movl a6@(20),a3"); 
    asm("movl a6@(16),a2"); 
    asm("movl a6@(12),a1");
#endif    
#endif

    return(run_lisp_func(address));
}

int call_lisp4(id, arg1, arg2, arg3, arg4)
int id, arg1, arg2, arg3, arg4;
{
    int address = fca(id);

#ifdef vax
    asm("movl 20(ap),r4");
    asm("movl 16(ap),r3");
    asm("movl 12(ap),r2");
    asm("movl  8(ap),r1");
#endif

#ifdef sun
#ifdef DREGMODEL
    asm("movl a6@(24),d3"); 
    asm("movl a6@(20),d2"); 
    asm("movl a6@(16),d1"); 
    asm("movl a6@(12),d0");
#endif
#ifdef AREGMODEL
    asm("movl a6@(24),a4"); 
    asm("movl a6@(20),a3"); 
    asm("movl a6@(16),a2"); 
    asm("movl a6@(12),a1");
#endif
#endif

    return(run_lisp_func(address));
}

int call_lisp5(id, arg1, arg2, arg3, arg4, arg5)
int id, arg1, arg2, arg3, arg4, arg5;
{
    int address = fca(id);

#ifdef vax
    asm("movl 24(ap),r5");
    asm("movl 20(ap),r4");
    asm("movl 16(ap),r3");
    asm("movl 12(ap),r2");
    asm("movl  8(ap),r1");
#endif

#ifdef sun
#ifdef DREGMODEL
    asm("movl a6@(28),d4"); 
    asm("movl a6@(24),d3"); 
    asm("movl a6@(20),d2"); 
    asm("movl a6@(16),d1"); 
    asm("movl a6@(12),d0");
#endif
#ifdef AREGMODEL
    asm("movl a6@(28),a5"); 
    asm("movl a6@(24),a4"); 
    asm("movl a6@(20),a3"); 
    asm("movl a6@(16),a2"); 
    asm("movl a6@(12),a1");
#endif
#endif

    return(run_lisp_func(address));
}

/* I you add more routines here to pass even more arguments, you
   must be aware of how many real regs the specific implementation
   uses, and pass anything greater than real regs in the argument block. */

/* End of file. */
